Unit myvalid;
{$F+}

interface

uses objects, validate;

 type
  PFRangeValidator = ^TFRangeValidator;
  TFRangeValidator = object(TFilterValidator)
    Min, Max: real;
    constructor Init(AMin, AMax: real);
    constructor Load(var S: TStream);
    procedure Error; virtual;
    function IsValid(const S: string): Boolean; virtual;
    procedure Store(var S: TStream);
    function Transfer(var S: String; Buffer: Pointer;
      Flag: TVTransfer): Word; virtual;
  end;

procedure RegisterMyValidate;

const
  RFRangeValidator: TStreamRec = (
    ObjType: 1082;
    VmtLink: Ofs(TypeOf(TFRangeValidator)^);
    Load: @TFRangeValidator.Load;
    Store: @TFRangeValidator.Store
  );


implementation

{$IFDEF Windows}
uses WinTypes, WinProcs, Strings, OWindows;
{$ELSE}
uses MsgBox;
{$ENDIF Windows}

constructor TFRangeValidator.Init(AMin, AMax: real);
begin
  inherited Init(['0'..'9','+','-','.','e','E']);
  Min := AMin;
  Max := AMax;
end;

constructor TFRangeValidator.Load(var S: TStream);
begin
  inherited Load(S);
  S.Read(Min, SizeOf(Max) + SizeOf(Min));
end;

{$IFDEF Windows}

procedure TFRangeValidator.Error;
var
  Params: array[0..1] of LongInt;
  MsgStr: array[0..80] of Char;
  s1,s2 : string;
begin
  Str (Min:0:2, s1); Str (Max:0:2, s2);
  Params[0] := LongInt(@s1);
  Params[1] := LongInt(@s2);
  wvsprintf(MsgStr, 'Real value must be in the range %s to %s.', Params);
  MessageBox(0, MsgStr, 'Validator', mb_IconExclamation or mb_Ok);
end;

{$ELSE}

procedure TFRangeValidator.Error;
var
  Params: array[0..1] of LongInt;
  s1,s2 : string;
begin
  Str (Min:0:2, s1); Str (Max:0:2,s2);
  Params[0] := LongInt(@s1);
  Params[1] := LongInt(@s2);
  MessageBox('Real value must be in the range %s to %s', @Params,
    mfError + mfOKButton);
end;

{$ENDIF Windows}

function TFRangeValidator.IsValid(const S: string): Boolean;
var
  Value: real;
  Code: Integer;
begin
  IsValid := False;
  if inherited IsValid(S) then
  begin
    Val(S, Value, Code);
    if (Code = 0) and (Value >= Min) and (Value <= Max) then
      IsValid := True;
  end;
end;

procedure TFRangeValidator.Store(var S: TStream);
begin
  inherited Store(S);
  S.Write(Min, SizeOf(Max) + SizeOf(Min));
end;

function TFRangeValidator.Transfer(var S: String; Buffer: Pointer;
  Flag: TVTransfer): Word;
var
  Value: real;
  Code: Integer;
begin
  if Options and voTransfer <> 0 then
  begin
    Transfer := SizeOf(Value);
    case Flag of
     vtGetData:
       begin
         Val(S, Value, Code);
         real(Buffer^) := Value;
       end;
     vtSetData:
       Str(real(Buffer^), S);
    end;
  end
  else
    Transfer := 0;
end;

procedure RegisterMyValidate;
begin
  RegisterType(RFRangeValidator);
end;

end.