(*-------------------------------------------------------------------------*)
(*                             NEFCLASS-PC                                 *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Crisp classification by a fuzzy perceptron with fuzzy antecedents and   *)
(* crisp conclusion weights                                                *)
(*+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* Filename: NFCL_DLG.PAS, unit for dialog windows                         *)
(* Programmer: Ulrike Nauck                                                *)
(* Last changes: 10/19/95                                                  *)
(*-------------------------------------------------------------------------*)
(* (c) Technical University of Braunschweig, 1994                          *)
(*-------------------------------------------------------------------------*)

unit nfcl_dlg;
interface
uses Objects, Drivers, App, Dialogs, Views, TextView, MsgBox,
     NFCL_Con, NFCL_Typ, NFCL_Obj;

type

 PMyListBox = ^TMyListBox;
 TMyListBox = object (TListBox)
  procedure FocusItem(Item: Integer); virtual;
 end;

 PText = ^Text;

 PControlParamDialog = ^TControlParamDialog;
 TControlParamDialog = object (TDialog)
     Epochs         : PInputLine;
     StopError      : PInputLine;
     NofError       : PInputLine;
     Update         : PInputLine;
     StopCriterion  : PCheckBoxes;
     ErrorEpochs    : PInputLine;
     constructor Init (var Bounds: TRect; ATitle: TTitleStr);
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
 end;

 PLearnParamDialog = ^TLearnParamDialog;
 TLearnParamDialog = object (TDialog)
     NoPassing      : PRadioButtons;
     Asymmetrical   : PRadioButtons;
     Intersect      : PRadioButtons;
     FixedConc      : PRadioButtons;
     UnitConc       : PRadioButtons;
     RuleCreation   : PRadioButtons;
     ReLearnRules   : PRadioButtons;
     sigma_a        : PInputLine;
     sigma_b        : PInputLine;
     sigma_c        : PInputLine;
     sigma          : PInputLine;
     constructor Init (var Bounds: TRect; ATitle: TTitleStr);
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
 end;

 PRControlParamBuf = ^TRControlParamBuf;
 TRControlParamBuf = record
  Epochs   : string[10];
  StopErr  : string[10];
  NofErrors: string[10];
  Update   : string[5];
  StopCriterion: Word;
  ErrorEpochs : string[6];
 end;

 PRLearnParamBuf = ^TRLearnParamBuf;
 TRLearnParamBuf = record
  NoPass      : Word;
  Asymmetrical: Word;
  Intersect : Word;
  FixedConc : Word;
  UnitConc  : Word;
  RuleCreation: Word;
  Relearn : Word;
  sigma_a  : string[10];
  sigma_b  : string[10];
  sigma_c  : string[10];
  sigma    : string[10];
 end;

 PPartList = ^TPartList;
 TPartList = object (TMyListBox)
  function GetText(Item: Integer; MaxLen: Integer): String; virtual;
  function GetFSet(Item: Integer; MaxLen: Integer): String;
 end;

 PNetParamDialog = ^TNetParamDialog;
 TNetParamDialog = object(TDialog)
     InLayer    : PInputLine;
     RuleLayer  : PInputLine;
     OutLayer   : PInputLine;
     UseSum     : PCluster;
     Individual : PCluster;
     Partition  : PInputLine;
     Picked     : PInputLine;
     IndFSets   : PInputLine;
     UnitList   : PPartList;
     ScrollBar  : PScrollBar;
     ClassPicked: PInputLine;
     ClassList  : PMyListBox;
     CScrollBar : PScrollBar;
     constructor Init (var Bounds: TRect; ATitle: TTitleStr);
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
     procedure HandleEvent(var Event: TEvent); virtual;
 end;

 PRNetParam = ^TRNetParam;
 TRNetParam = record
   InLayer    : string[10];
   RuleLayer  : string[10];
   OutLayer   : string[10];
   UseSum     : word;
   Individual : word;
   Partition  : string[5];
   Picked     : string[9];
   IndFSets   : string[5];
   ListPtr    : PPartCollection;
   Focussed   : word;
   ClassPicked: string[9];
   CListPtr   : PMyStrCollection;
   CFocussed  : word;
 end;

 PTermWin = ^TTermWin;
 TTermWin = object (TWindow)
  Term    : Text;
  Terminal: PTerminal;
  constructor Init (ATitle: TTitleStr; W, H: Word);
  procedure GetTerm (var ATerm: PText);
 end;

 PPlotDialog = ^TPlotDialog;
 TPlotDialog = object (TDialog)
     V1         : PInputLine;
     V2         : PInputLine;
     PlotSelect : PRadioButtons;
     constructor Init (var Bounds: TRect; ATitle: TTitleStr);
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
 end;

 PRPlotParam = ^TRPlotParam;
 TRPlotParam = record
   V1           : string[10];
   V2           : string[10];
   PlotSelect   : word;
 end;

 PRuleList = ^TRuleList;
 TRuleList = object (TMyListBox)
  function GetText(Item: Integer; MaxLen: Integer): String; virtual;
  function GetClass(Item: Integer): String;
  function GetAntec(Item: Integer; Part: Integer): String;
 end;

 PRuleDialog = ^TRuleDialog;
 TRuleDialog = object(TDialog)
     AntecList  : PListBox;
     AScrollBar : PScrollBar;
     Class      : PInputLine;
     RuleList   : PRuleList;
     RScrollBar : PScrollBar;
     FSetList   : PListBox;
     FScrollBar : PScrollBar;
     VarList    : PMyListBox;
     VScrollBar : PScrollBar;
     ClassList  : PListBox;
     CScrollBar : PScrollBar;
     Net        : PNefclass;
     constructor Init (var Bounds: TRect; ATitle: TTitleStr);
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
     function  Valid(Command: Word): Boolean; virtual;
     function  TestRules(Command: Word): Boolean;
     procedure NewAntecList(Rule: integer);
     procedure CreateNewRule;
     procedure RenumberRules (Start: integer);
     procedure DeleteRule;
     procedure EmptyRule;
     procedure FocusRule;
     procedure NewFuzzyList (V: integer);
     procedure ModifyAntec;
     procedure AddAntec (Modify: Boolean);
     procedure DeleteAntec;
     procedure SetClass;
     procedure HandleEvent(var Event: TEvent); virtual;
     procedure InitData (ANet: PNefclass);
 end;

 PRRuleParam = ^TRRuleParam;
 TRRuleParam = record
   AListPtr   : PAntStrCollection;
   AFocussed  : word;
   Class      : string[10];
   RListPtr   : PRuleStrCollection;
   RFocussed  : word;
   FListPtr   : PMyStrCollection;
   FFocussed  : word;
   VListPtr   : PMyStrCollection;
   VFocussed  : word;
   CListPtr   : PMyStrCollection;
   CFocussed  : word;
 end;

 PNetLabelDialog = ^TNetLabelDialog;
 TNetLabelDialog = object(TDialog)
     NetLabel  : PInputLine;
     constructor Load (var S: TStream);
     procedure Store (var S: TStream); virtual;
 end;

const
  RLearnParamDialog: TStreamRec = (
    ObjType: 1001;
    VmtLink: Ofs(TypeOf(TLearnParamDialog)^);
    Load: @TLearnParamDialog.Load;
    Store: @TLearnParamDialog.Store
  );

  RNetParamDialog: TStreamRec = (
    ObjType: 1002;
    VmtLink: Ofs(TypeOf(TNetParamDialog)^);
    Load: @TNetParamDialog.Load;
    Store: @TNetParamDialog.Store
  );

  RControlParamDialog: TStreamRec = (
    ObjType: 1003;
    VmtLink: Ofs(TypeOf(TControlParamDialog)^);
    Load: @TControlParamDialog.Load;
    Store: @TControlParamDialog.Store
  );

  RPlotDialog: TStreamRec = (
    ObjType: 1004;
    VmtLink: Ofs(TypeOf(TPlotDialog)^);
    Load: @TPlotDialog.Load;
    Store: @TPlotDialog.Store
  );

  RPartList: TStreamRec = (
    ObjType: 1005;
    VmtLink: Ofs(TypeOf(TPartList)^);
    Load: @TPartList.Load;
    Store: @TPartList.Store
  );

  RRuleDialog: TStreamRec = (
    ObjType: 1006;
    VmtLink: Ofs(TypeOf(TRuleDialog)^);
    Load: @TRuleDialog.Load;
    Store: @TRuleDialog.Store
  );

  RRuleList: TStreamRec = (
    ObjType: 1007;
    VmtLink: Ofs(TypeOf(TRuleList)^);
    Load: @TRuleList.Load;
    Store: @TRuleList.Store
  );

  RMyListBox: TStreamRec = (
    ObjType: 1008;
    VmtLink: Ofs(TypeOf(TMyListBox)^);
    Load: @TMyListBox.Load;
    Store: @TMyListBox.Store
  );

  RNetLabelDialog: TStreamRec = (
    ObjType: 1009;
    VmtLink: Ofs(TypeOf(TNetLabelDialog)^);
    Load: @TNetLabelDialog.Load;
    Store: @TNetLabelDialog.Store
  );

var
 ResFile       : TResourceFile;

procedure RegisterMyDialogs;

implementation

(*-------------------------------------------------------------------------*)
(*                           Methods of TMyListBox                         *)
(*-------------------------------------------------------------------------*)

procedure TMyListBox.FocusItem (Item: integer);
 begin
  inherited FocusItem(Item);
  if Range > Focused then SelectItem(Item);
 end;

(*-------------------------------------------------------------------------*)
(*                        Methods of TControlParamDialog                   *)
(*-------------------------------------------------------------------------*)

constructor TControlParamDialog.Init (var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init(Bounds, ATitle);
 end;

constructor TControlParamDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, Epochs);
  GetSubViewPtr (S, StopError);
  GetSubViewPtr (S, NofError);
  GetSubViewPtr (S, Update);
  GetSubViewPtr (S, StopCriterion);
  GetSubViewPtr (S, ErrorEpochs);
 end; (* TControlParamDialog.Load *)

procedure TControlParamDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, Epochs);
  PutSubViewPtr (S, StopError);
  PutSubViewPtr (S, NofError);
  PutSubViewPtr (S, Update);
  PutSubViewPtr (S, StopCriterion);
  PutSubViewPtr (S, ErrorEpochs);
 end; (* TControlParamDialog.Store *)

(*-------------------------------------------------------------------------*)
(*                        Methods of TLearnParamDialog                     *)
(*-------------------------------------------------------------------------*)

constructor TLearnParamDialog.Init (var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init(Bounds, ATitle);
 end;

constructor TLearnParamDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, sigma_a);
  GetSubViewPtr (S, sigma_b);
  GetSubViewPtr (S, sigma_c);
  GetSubViewPtr (S, sigma);
  GetSubViewPtr (S, NoPassing);
  GetSubViewPtr (S, Asymmetrical);
  GetSubViewPtr (S, Intersect);
  GetSubViewPtr (S, FixedConc);
  GetSubViewPtr (S, UnitConc);
  GetSubViewPtr (S, RuleCreation);
  GetSubViewPtr (S, ReLearnRules);
 end; (* TLearnParamDialog.Load *)

procedure TLearnParamDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, sigma_a);
  PutSubViewPtr (S, sigma_b);
  PutSubViewPtr (S, sigma_c);
  PutSubViewPtr (S, sigma);
  PutSubViewPtr (S, NoPassing);
  PutSubViewPtr (S, Asymmetrical);
  PutSubViewPtr (S, Intersect);
  PutSubViewPtr (S, FixedConc);
  PutSubViewPtr (S, UnitConc);
  PutSubViewPtr (S, RuleCreation);
  PutSubViewPtr (S, ReLearnRules);
 end; (* TLearnParamDialog.Store *)

(*-------------------------------------------------------------------------*)
(*                           Methods of TPartList                          *)
(*-------------------------------------------------------------------------*)

function TPartList.GetText(Item: Integer; MaxLen: Integer): String;
 var
  APartition : PPartition;
 begin
  if List <> nil
    then begin
     APartition := PPartition(List^.At(Item));
     if APartition <> nil
       then GetText := Copy (APartition^.VName, 1, MaxLen)
       else GetText := '';
    end
    else GetText := '';
 end;

function TPartList.GetFSet(Item: Integer; MaxLen: Integer): String;
 var
  APartition : PPartition;
  S          : string[5];
 begin
  if List <> nil
    then begin
     APartition := PPartition(List^.At(Item));
     if APartition <> nil
       then begin
        Str (APartition^.Count, S);
        GetFSet := Copy (S, 1, MaxLen);
       end
       else GetFSet := '0';
    end
    else GetFSet := '0';
 end;

(*-------------------------------------------------------------------------*)
(*                        Methods of TNetParamDialog                       *)
(*-------------------------------------------------------------------------*)

constructor TNetParamDialog.Init (var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init (Bounds, ATitle);
 end; (* TNetParamDialog.Init *)

constructor TNetParamDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, InLayer);
  GetSubViewPtr (S, RuleLayer);
  GetSubViewPtr (S, OutLayer);
  GetSubViewPtr (S, UseSum);
  GetSubViewPtr (S, Partition);
  GetSubViewPtr (S, Individual);
  GetSubViewPtr (S, Picked);
  GetSubViewPtr (S, IndFSets);
  GetSubViewPtr (S, UnitList);
  GetSubViewPtr (S, ScrollBar);
  GetSubViewPtr (S, ClassPicked);
  GetSubViewPtr (S, ClassList);
  GetSubViewPtr (S, CScrollBar);
 end; (* TNetParamDialog.Init *)

procedure TNetParamDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, InLayer);
  PutSubViewPtr (S, RuleLayer);
  PutSubViewPtr (S, OutLayer);
  PutSubViewPtr (S, UseSum);
  PutSubViewPtr (S, Partition);
  PutSubViewPtr (S, Individual);
  PutSubViewPtr (S, Picked);
  PutSubViewPtr (S, IndFSets);
  PutSubViewPtr (S, UnitList);
  PutSubViewPtr (S, ScrollBar);
  PutSubViewPtr (S, ClassPicked);
  PutSubViewPtr (S, ClassList);
  PutSubViewPtr (S, CScrollBar);
 end; (* TNetParamDialog.Store *)

procedure TNetParamDialog.HandleEvent(var Event: TEvent);
 var
  Value, Code : integer;
 begin
   inherited HandleEvent(Event);
   if (Event.What = evCommand) or (Event.What = evBroadcast)
    then begin
     case Event.Command of
      cmFSet : begin
                PPartition(UnitList^.List^.At(UnitList^.Focused))^.VName :=
                Picked^.Data^;
                if IndFSets^.Valid(cmClose)
                 then begin
                  Val (IndFSets^.Data^, Value, Code);
                  PPartition(UnitList^.List^.At(UnitList^.Focused))^.Count :=
                  Value;
                 end;
                UnitList^.Draw;
               end;
      cmCSet : begin
                PVNStr(ClassList^.List^.At(ClassList^.Focused))^ :=
                ClassPicked^.Data^;
                ClassList^.Draw;
               end;
      cmListItemSelected :
               begin
                 if Event.InfoPtr = ClassList
                  then begin
                   ClassPicked^.Data^ :=
                    ClassList^.GetText(ClassList^.Focused, 9);
                   ClassPicked^.DrawView;
                  end
                  else if Event.InfoPtr = UnitList
                   then begin
                    Picked^.Data^ :=
                     UnitList^.GetText(UnitList^.Focused, 9);
                    IndFSets^.Data^ :=
                     UnitList^.GetFSet(UnitList^.Focused, 5);
                    IndFSets^.DrawView;
                    Picked^.DrawView;
                   end;
               end;
     end; (* case *)
     DrawView;
     ClearEvent(Event);
    end;
 end;


(*-------------------------------------------------------------------------*)
(*                          Methods of TTermWin                            *)
(*-------------------------------------------------------------------------*)

constructor TTermWin.Init (ATitle: TTitleStr; W, H: Word);
 var
  HScrollbar : PScrollbar;
  VScrollbar : PScrollbar;
  R          : TRect;
 begin
  R.Assign (0,0,W,H);
  inherited init(R, ATitle, wnNoNumber);
  R.Grow (-1, -1);
  HScrollbar := StandardScrollBar(sbHorizontal or sbHandleKeyboard);
  Insert (HScrollbar);
  VScrollbar := StandardScrollBar(sbVertical or sbHandleKeyboard);
  Insert (VScrollbar);
  New(Terminal, Init(R, HScrollbar, VScrollbar, 8192));
  If Application^.ValidView(Terminal) <> nil
    then begin
     AssignDevice (Term, Terminal);
     Rewrite (Term);
     Insert (Terminal);
    end;
 end; (* TTermWin *)

procedure TTermWin.GetTerm (var ATerm: PText);
 begin
  ATerm := @Term;
 end; (* TTermWin.GetText *)

(*-------------------------------------------------------------------------*)
(*                          Methods of TPlotDialog                         *)
(*-------------------------------------------------------------------------*)

constructor TPlotDialog.Init (var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init(Bounds, ATitle);
 end;

constructor TPlotDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, V1);
  GetSubViewPtr (S, V2);
  GetSubViewPtr (S, PlotSelect);
 end; (* TPlotDialog.Load *)

procedure TPlotDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, V1);
  PutSubViewPtr (S, V2);
  PutSubViewPtr (S, PlotSelect);
 end; (* TPlotDialog.Store *)

(*-------------------------------------------------------------------------*)
(*                           Methods of TPartList                          *)
(*-------------------------------------------------------------------------*)

function TRuleList.GetText(Item: Integer; MaxLen: Integer): String;
 var
  ARule : PRuleDescriptor;
 begin
  if List <> nil
    then begin
     ARule := PRuleDescriptor(List^.At(Item));
     if ARule <> nil
       then GetText := Copy (ARule^.Name, 1, MaxLen)
       else GetText := '';
    end
    else GetText := '';
 end; (* TRuleList.GetText *)

(*-------------------------------------------------------------------------*)

function TRuleList.GetClass(Item: Integer): String;
 var
  ARule : PRuleDescriptor;
 begin
  if List <> nil
    then begin
     ARule := PRuleDescriptor(List^.At(Item));
     if ARule <> nil
       then begin
        GetClass := ARule^.Conclusion;
       end
       else GetClass := '';
    end
    else GetClass := '';
 end;

(*-------------------------------------------------------------------------*)

function TRuleList.GetAntec(Item: Integer; Part: Integer): String;
 const
  blanks = '        ';
 var
  ARule : PRuleDescriptor;
  C: string[4];
  S: AntStr;
 begin
  if List <> nil
    then begin
     ARule := PRuleDescriptor(List^.At(Item));
     if ARule <> nil
       then begin
        S := ARule^.Antecedent[Part].Variable;
        if S <> ''
          then begin
           GetAntec := S + Copy (blanks, 1, 8 - length(S)) + ' IS ' +
                       ARule^.Antecedent[Part].Value;
          end
          else GetAntec := '';
       end
       else GetAntec := '';
    end
    else GetAntec := '';
 end;

(*-------------------------------------------------------------------------*)
(*                        Methods of TRuleDialog                           *)
(*-------------------------------------------------------------------------*)

constructor TRuleDialog.Init (var Bounds: TRect; ATitle: TTitleStr);
 begin
  inherited Init (Bounds, ATitle);
 end; (* TNetParamDialog.Init *)

(*-------------------------------------------------------------------------*)

constructor TRuleDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, AntecList);
  GetSubViewPtr (S, AScrollBar);
  GetSubViewPtr (S, Class);
  GetSubViewPtr (S, RuleList);
  GetSubViewPtr (S, RScrollBar);
  GetSubViewPtr (S, FSetList);
  GetSubViewPtr (S, FScrollBar);
  GetSubViewPtr (S, VarList);
  GetSubViewPtr (S, VScrollBar);
  GetSubViewPtr (S, ClassList);
  GetSubViewPtr (S, CScrollBar);
 end; (* TRuleDialog.Init *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, AntecList);
  PutSubViewPtr (S, AScrollBar);
  PutSubViewPtr (S, Class);
  PutSubViewPtr (S, RuleList);
  PutSubViewPtr (S, RScrollBar);
  PutSubViewPtr (S, FSetList);
  PutSubViewPtr (S, FScrollBar);
  PutSubViewPtr (S, VarList);
  PutSubViewPtr (S, VScrollBar);
  PutSubViewPtr (S, ClassList);
  PutSubViewPtr (S, CScrollBar);
 end; (* TRuleDialog.Store *)

(*-------------------------------------------------------------------------*)

function  TRuleDialog.Valid(Command: Word): Boolean;
 var
  Result : Boolean;
 begin
  if (Command <> cmValid) and (* test rules only, if the dialog is neither *)
     (Command <> cmCancel)    (* checked after initialization, nor canceled*)
    then begin
     Result := TestRules (Command);
     if Result then Result := inherited Valid (Command);
    end
    else Result := inherited Valid (Command);
  Valid := Result;
 end;

(*-------------------------------------------------------------------------*)
(* The method tests the rule base for consistency. This is automatically   *)
(* done, when the dialog is closed via the OK button, and it can also be   *)
(* used to check, wether the recently included rule fits into the rule     *)
(* base. If Command = cmOK, then the whole rule base is checked, if not,   *)
(* only the last rule of the rule list is tested.                          *)
(* It returns true, if the rule base is consistent.                        *)
(*-------------------------------------------------------------------------*)

function  TRuleDialog.TestRules(Command: Word): Boolean;
 var
  identical     : Boolean;
  superset      : Boolean;
  subset        : Boolean;
  contradictory : Boolean;
  NoClass       : Boolean;
  OK            : Boolean;
  BoxText       : string;
  CurrentRule   : PRuleDescriptor;
  ErrorRule     : PRuleDescriptor;
  i             : integer;

 (*--------------------------------------------------------------*)
 (* This local function compares two rules, and returns true, if *)
 (* they are identical or if one is a generalization or a        *)
 (* specialization of the other one.                             *)
 (*--------------------------------------------------------------*)
 function CompareRules (Rule1, Rule2: PRuleDescriptor): Boolean;
  var
   Result   : Boolean;
   varident : Boolean;
   i        : integer;
  begin
   identical     := true;
   superset      := true;
   subset        := true;
   contradictory := false;
   i := 1;
   while (i <= Net^.L1) and (identical or superset or subset)
    do begin
     varident := (Rule1^.Antecedent[i].Value = Rule2^.Antecedent[i].Value);
     identical := identical and varident;
     if not identical
       then begin
        superset := superset and (varident or
                                  (Rule1^.Antecedent[i].Value = ''));
        subset := subset and (varident or (Rule2^.Antecedent[i].Value = ''));
       end;
     inc (i);
    end; (* while *)
    Result := identical or (not identical and (superset or subset));
    if Result
      then begin
       contradictory := (Rule1^.Conclusion <> Rule2^.Conclusion);
      end;
   CompareRules := Result;
  end; (* CompareRules *)

 (*----------------------------------------------------------------*)
 (* This local function checks the rule given in ARule against the *)
 (* rules in the rule base starting at the position given in from, *)
 (* and stopping at the position given in stop.                    *)
 (* It returns False, if the rule does not fit into the rule base. *)
 (*----------------------------------------------------------------*)

 function CheckRule (ARule: PRuleDescriptor; from, stop: integer): Boolean;
  var
   Error: Boolean;
  begin
   if ARule^.Conclusion = ''
     then begin
      Error := True;
      NoClass := True;
     end
     else begin
      Error := False;
      NoClass := False;
     end;
   while not Error and (from <= stop)
    do begin
     Error := CompareRules (ARule, RuleList^.List^.At(from));
     inc (from);
    end;
   if Error
     then begin
      if NoClass
        then ErrorRule := nil
        else ErrorRule := RuleList^.List^.At(from-1); (* conflicting rule *)
      CheckRule := False;
     end
     else begin
      ErrorRule := nil;
      CheckRule := True;
     end;
  end; (* CheckRule *)

 (* ------------------ main routine of method TestRules --------------- *)

 begin
  identical     := False;
  superset      := False;
  subset        := False;
  contradictory := False;
  NoClass       := False;

  if Command = cmOK (* is window to be closed? *)
    then begin (* test all rules *)
     i := 0; OK := True;
     while OK and (i < RuleList^.List^.Count)
      do begin
       CurrentRule := RuleList^.List^.At(i);
       OK := CheckRule (CurrentRule, i+1, RuleList^.List^.Count-1);
       inc (i);
      end;
    end
    else begin (* test only the last rule in the list *)
     CurrentRule := RuleList^.List^.At(RuleList^.Focused);
     OK := CheckRule (CurrentRule, 0, RuleList^.Focused-1);
    end;
  if not OK
    then begin
     BoxText := #3'The rule ';
     if Command = cmOK
       then BoxText := BoxText + CurrentRule^.Name
       else BoxText := BoxText + 'that you tried to enter';
     BoxText := BoxText + #13#3;
     if identical
       then BoxText := BoxText + 'is identical to rule '
       else if superset
              then BoxText := BoxText + 'is a generalization of rule '
              else if subset
                then BoxText := BoxText + 'is a specialization of rule '
                else if NoClass
                  then BoxText := BoxText + 'has no conclusion!'#13#3
                  else BoxText := BoxText +
                                  'caused an unexpected error!'#13#3;
     if not NoClass
       then BoxText := BoxText + ErrorRule^.Name + ',' + #13#3;
     if contradictory
       then BoxText := BoxText + 'and causes a contradiction!'#13#3
       else if not NoClass
         then BoxText := BoxText + 'and therfore one is superfluous!'#13#3;
     if Command = cmOK
       then BoxText := BoxText + 'You must correct the rule base!'
       else begin
        BoxText := BoxText + 'Delete or change this rule!';
       end;
     MessageBox (BoxText, nil, mfError or mfOKButton);
     TestRules := False;
    end
    else TestRules := True;
 end; (* TRuleDialog.TestRules *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.NewAntecList (Rule: integer);
 var
  S   : AntStr;
  Name: PAntStr;
  i: integer;
  TheFirst: Boolean;
 begin
   if AntecList^.List <> nil
     then begin
      AntecList^.List^.FreeAll;
      TheFirst := True;
      for i := 1 to Net^.L1 do
       begin
        S := RuleList^.GetAntec(Rule,i);
        if S <> ''
         then begin
          if TheFirst
           then begin
            S := '    ' + S;
            TheFirst := False;
           end
           else S := 'AND ' + S;
          new (Name);
          Name^ := S;
          AntecList^.List^.Insert (Name);
         end;
       end;
      AntecList^.Range := AntecList^.List^.Count;
      AntecList^.VScrollBar^.SetParams (0,0,AntecList^.Range,1,1);
      if AntecList^.Range > 0 then AntecList^.FocusItem(0);
      AntecList^.DrawView;
     end;
 end;

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.RenumberRules (Start: integer);
 var
  S: VNStr;
  i: integer;
 begin
  for i := Start to RuleList^.List^.Count do
   begin
    Str (i, S);
    PRuleDescriptor(RuleList^.List^.At(i-1))^.Name := 'R_' + S;
   end;

 end; (* TRuleDialog.RenumberRules *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.CreateNewRule;
 var
  Rule: PRuleDescriptor;
  S   : VNStr;
  i   : integer;
 begin
  if RuleList^.Range >= RuleLimit
    then MessageBox ('You have reached the maximum number of rules. '+
                     'Sorry, but you cannot enter more rules!', nil,
                     mfOKButton)
    else begin
     new (Rule);
     Str (RuleList^.Range + 1, S);
     Rule^.Name := 'R_' + S;
     Rule^.Conclusion := '';
     for i := 1 to Net^.L1 do
      begin
       Rule^.Antecedent[i].Variable := '';
       Rule^.Antecedent[i].Value := '';
      end;
     RuleList^.List^.Insert (Rule);
     RuleList^.Range := RuleList^.List^.Count;
     RuleList^.VScrollBar^.SetParams (0,0,RuleList^.Range,1,1);
     RuleList^.FocusItem(RuleList^.Range - 1);
     RuleList^.DrawView;
     FocusRule;
     if RuleList^.Range = Net^.max_rules + 1
       then MessageBox ('Be careful! You now have created more rules '+
                        'then your network is supposed to have. You may '+
                        'want to change the network parameters.', nil,
                        mfOKButton);
    end; (* else *)
 end; (* TRuleDialog.CreateNewRule *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.DeleteRule;
 var
  Rule: PRuleDescriptor;
  S   : VNStr;
  i   : integer;
 begin
  i := RuleList^.Focused;
  RuleList^.List^.AtFree (i);
  RuleList^.Range := RuleList^.List^.Count;
  RuleList^.VScrollBar^.SetParams (0, 0, RuleList^.Range, 1, 1);
  if i > RuleList^.Range-1
    then i := RuleList^.Range-1;
  if RuleList^.Range > 0
    then begin
     RenumberRules (i+1);
     RuleList^.FocusItem(i);
     FocusRule;
    end
    else EmptyRule;

  RuleList^.DrawView;
 end; (* TRuleDialog.DeleteRule *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.EmptyRule;
 begin
  Class^.Data^ := '';
  Class^.DrawView;
  AntecList^.List^.FreeAll;
  AntecList^.Range := 0;
  AntecList^.VScrollBar^.SetParams (0,0,0,1,1);
  AntecList^.DrawView;
 end; (* TRuleDialog.EmptyRule *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.FocusRule;
 begin
  Class^.Data^ := RuleList^.GetClass(RuleList^.Focused);
  Class^.DrawView;
  NewAntecList(RuleList^.Focused);
  AntecList^.DrawView;
 end; (* TRuleDialog.FocusRule *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.NewFuzzyList (V: integer);
 var
  S   : AntStr;
  Name: PAntStr;
  i: integer;
 begin
   if FSetList^.List <> nil
     then begin
      FSetList^.List^.FreeAll;
      for i := 1 to Net^.AntecBase[V].count do
       begin
        new (Name);
        Name^ := Net^.AntecBase[V].FSets[i].Name;
        FSetList^.List^.Insert (Name);
       end;
      FSetList^.Range := FSetList^.List^.Count;
      FSetList^.VScrollBar^.Max := FSetList^.Range;
      if FSetList^.Range > 0 then FSetList^.FocusItem(0);
      FSetList^.DrawView;
     end;
 end; (* TRuleDialog.NewFuzzyList *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.ModifyAntec;
 var
  Rule: PRuleDescriptor;
 begin
  Rule := PRuleDescriptor(RuleList^.List^.At(RuleList^.Focused));
  if Rule^.Antecedent[VarList^.Focused+1].Variable <> ''
    then AddAntec (True)
    else MessageBox (#3'This variable is currently not included '+
                     'in the seleceted rule!',nil, mfError or mfOKButton);
 end; (* ModifyAntec *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.AddAntec (Modify: Boolean);
 var
  Rule: PRuleDescriptor;

 begin
  Rule := PRuleDescriptor(RuleList^.List^.At(RuleList^.Focused));
  if (Rule^.Antecedent[VarList^.Focused+1].Variable = '') or Modify
    then begin
     Rule^.Antecedent[VarList^.Focused+1].Variable :=
        VarList^.GetText(VarList^.Focused, 8);
     Rule^.Antecedent[VarList^.Focused+1].Value :=
        FSetList^.GetText(FSetList^.Focused, 8);
     FocusRule;
    end
    else MessageBox (#3'This variable is already included '+
                     'in the seleceted rule!',nil, mfError or mfOKButton);
 end; (* TRuleDialog.AddAntec *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.DeleteAntec;
 var
  Rule: PRuleDescriptor;

 begin
  Rule := PRuleDescriptor(RuleList^.List^.At(RuleList^.Focused));
  if Rule^.Antecedent[VarList^.Focused+1].Variable <> ''
    then begin
     Rule^.Antecedent[VarList^.Focused+1].Variable := '';
     Rule^.Antecedent[VarList^.Focused+1].Value := '';
     FocusRule;
    end
    else MessageBox (#3'This variable is currently not included '+
                     'in the seleceted rule!',nil, mfError or mfOKButton);
 end; (* TRuleDialog.DeleteAntec *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.SetClass;
 var
  Rule: PRuleDescriptor;

 begin
  Rule := PRuleDescriptor(RuleList^.List^.At(RuleList^.Focused));
  Rule^.Conclusion := ClassList^.GetText(ClassList^.Focused,8);
  Class^.Data^ := RuleList^.GetClass(RuleList^.Focused);
  Class^.DrawView;
  TestRules (cmValid);
 end; (* TRuleDialog.SetClass *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.HandleEvent(var Event: TEvent);
 const
  AddIt    = False;
  ModifyIt = True;
 var
  Value, Code : integer;
 begin
   inherited HandleEvent(Event);
   if (Event.What = evCommand) or (Event.What = evBroadcast)
    then begin
     case Event.Command of
      cmEnterRule : CreateNewRule;
      cmDeleteRule: if RuleList^.Range > 0 then DeleteRule;
      cmEnterAntec: if RuleList^.Range > 0 then AddAntec(AddIt);
      cmModifyAntec: if RuleList^.Range > 0 then ModifyAntec;
      cmDeleteAntec: if RuleList^.Range > 0 then DeleteAntec;
      cmSetRuleClass: if RuleList^.Range > 0 then SetClass;
      cmListItemSelected : begin
                            if Event.InfoPtr = RuleList
                              then FocusRule
                            else if Event.InfoPtr = VarList
                              then NewFuzzyList (VarList^.Focused + 1);
                           end;
     end; (* case *)
     DrawView;
     ClearEvent(Event);
    end;
 end; (* TRuleDialog.HandleEvent *)

(*-------------------------------------------------------------------------*)

procedure TRuleDialog.InitData (ANet: PNefclass);
 begin
  Net := ANet;
 end;

(*-------------------------------------------------------------------------*)
(*                        Methods of TNetLabelDialog                       *)
(*-------------------------------------------------------------------------*)

constructor TNetLabelDialog.Load (var S: TStream);
 begin
  inherited Load (S);
  GetSubViewPtr (S, NetLabel);
 end;

(*-------------------------------------------------------------------------*)

procedure TNetLabelDialog.Store (var S: TStream);
 begin
  inherited Store (S);
  PutSubViewPtr (S, NetLabel);
 end;

(*-------------------------------------------------------------------------*)

procedure RegisterMyDialogs;
 begin
  RegisterType (RLearnParamDialog);
  RegisterType (RControlParamDialog);
  RegisterType (RNetParamDialog);
  RegisterType (RPlotDialog);
  RegisterType (RPartList);
  RegisterType (RRuleDialog);
  RegisterType (RRuleList);
  RegisterType (RMyListBox);
  RegisterType (RNetLabelDialog);
 end;

end.