(***************************************************************************)
(* (c) Detlef Nauck, 22.April 1992                                         *)
(*-------------------------------------------------------------------------*)
(*                          NEURO-FUZZY-LEARN                              *)
(*-------------------------------------------------------------------------*)
(* Ein Programm zur Simulation eines lernenden Fuzzy Controllers am        *)
(* Beispiel des inversen Pendels                                           *)
(***************************************************************************)

Program NFCLearn;
Uses Objects, NFCtrl, NPendSim, Graph, Graphics, Crt, Dos;

Var
 Controller : PFuzzyController;
 Pendulum   : PPendel;
 Loops, I,J : LongInt;
 Error      : Real;
 C_Error    : Real;
 ErrorAll   : Real;
 ErrorOld   : Real;
 Exit       : Boolean;
 NewTrial   : Boolean;
 Dummy      : Char;
 Time       : Integer;
 Starts     : Integer;
 Index      : Integer;
 StartW     : Real;
 StartWG    : Real;
 RandStartW : Real;
 RandStartWG: Real;
 Result     : Text;
 Rulefile   : Text;
 Year, Month, Day, DayOfWeek: Word;
 Hour, Minute, Second, Sec100: Word;
 FWindow: FuzzyWindow;
 GrDriver, GrMode: Integer;

Begin
 GetDate(Year, Month, Day, DayOfWeek);
 GetTime(Hour, Minute, Second, Sec100);
 VarFileName := 'NFC2.VAR';
 NonLin := 3;
 MFSLearn := false;

 Controller:= New(PFuzzyController, Init ('MeinController'));
 Pendulum  := New(PPendel, Init(SystemInState, OldSystemInState,
                                SystemOutState, OldSystemOutState));

 Assign (Result, 'NEUROFUZ.VAR');
 Rewrite (Result);
 Writeln (Result, Comment);
 Writeln (Result);
 Writeln (Result, 'Zustand zu Simulationsbeginn'); Writeln (Result);
 PrintFuzzyVariables (Result);
 Assign (Rulefile, 'NEUROFUZ.RUL');
 Rewrite (Rulefile);
 Writeln (Rulefile, 'Versuch vom: ', Day,'.',Month,'.',Year, ' um ',
          Hour,':',Minute,':',Second);
 Writeln (Rulefile);
 Assign (KillLog, 'NEUROFUZ.KIL');
 Rewrite (KillLog);
 KillLogging := True;

 Repeat
  
  ClrScr;
  Writeln (' ':16, 'Lernender Fuzzy-Controller mit Pendelsimulation');
  Writeln (' ':26, '(c) Detlef Nauck, April 1992');
  Window (20,8,60,15);
  Write ('Startwinkel               : ');
  Readln (StartW);
  Write ('Startwinkelgeschwindigkeit: ');
  Readln (StartWG);

  SetInState (SystemInState,'Angle',StartW);
  SetInState (SystemInState,'AngleVelocity', StartWG);
  SystemOutState^.ChangeState(0);
  SetInState (OldSystemInState,'Angle',StartW);
  SetInState (OldSystemInState,'AngleVelocity', StartWG);
  OldSystemOutState^.ChangeState(0);

  Write ('Anzahl der Zeitscheiben   : ');
  Readln (Loops);
  Write ('Verzgerung in ms         : ');
  Readln (Time);
  Time := Time - (Time Mod 10);
  Repeat
   ClrEol;
   Write ('Lernen ? (J/N)            : ');
   Readln  (Dummy);
   GotoXY (1,WhereY-1);
   Dummy := UpCase (Dummy);
  Until Dummy in ['J','N'];
  Learning := Dummy = 'J';
  GotoXY (1,WhereY+1);
  Repeat
   ClrEol;
   Write ('nderungsprotokoll? (J/N) : ');
   Readln  (Dummy);
   GotoXY (1,WhereY-1);
   Dummy := UpCase (Dummy);
  Until Dummy in ['J','N'];
  Logging := Dummy = 'J';
  
  ClrScr;

  Window  (25,4,55,7);
  Writeln ('        Anfangszustand');
  Writeln ('Winkel               : ', StartW:6:2);
  Writeln ('Winkelgeschwindigkeit: ', StartWG:6:2);
  Write   ('Anzahl Zeitscheiben  : ', Loops:6);

  Window (25,9,55,9);
  Write   ('       Aktueller Zustand');
  Window (25,10,57,21);
  Writeln ('Winkel               :        ');
  Writeln ('Winkelgeschwindigkeit:         ');
  Writeln ('Kraft                :        N');
  Writeln ('Fehler               :         ');
  Writeln ('Fehlernderung       :         ');
  Writeln ('berschwing-Fehler   :         ');
  Writeln ('Gesamtfehler         :         ');
  Writeln ('Durchlauf            :         ');
  Writeln ('Versuch              :         ');
  Writeln ('Verzgerung          :        ms');
  Writeln ('Lernen               :          ');
  Write   ('Anzahl der Regeln    :          ');
  Window (48,10,54,21);

  Error := 0; ErrorOld := 0; ErrorAll := 0; C_Error := 0;
  I := 1; Exit := False; Starts := 1; NewTrial := True; J:= 1;
  If Logging
    Then Begin
     Assign (Log, 'FUZZY.LOG');
     Rewrite (Log)
    End;

  While (I <= Loops) And Not Exit Do
   Begin
    ClrScr;
    Writeln (GetInState ('Angle'):6:2);
    Writeln (GetInState ('AngleVelocity'):6:2);
    Writeln (SystemOutState^.Value:6:2);
    Writeln (Error:6:2);
    Writeln (Error - ErrorOld:6:2);
    Writeln (C_Error:6:2);
    Writeln ((ErrorAll / J):6:2);
    Writeln (I:6);
    Writeln (Starts:6);
    Writeln (Time:6);
    If NewTrial
      Then Begin
       NewTrial := False
      End;
    If Learning
      Then Writeln ('Ja':6)
      Else Writeln ('Nein':6);
    Write (Controller^.NofRules:6);

    Controller^.DetermineControlAction;
    Pendulum^.DetermineState;
    ErrorAll := ErrorAll + Abs (GetInState('Angle'));
    ErrorOld := Error;

    Controller^.Learn (Error, C_Error);
    If Learning and KillPhase 
      Then Begin
       If I = 2 * MaxCredit Then Controller^.FindBestRules;
       Controller^.KillRules(I);
      End;

    If KeyPressed
      Then Begin
       While KeyPressed Do
        Dummy := ReadKey;
       Case Dummy of
          #13: Learning := (Not Learning);
          #27: Exit := True;
          #72: Begin
                If Time < 5000 Then Inc (Time,10)
               End;
          #80:Begin
               If Time > 0 Then Dec (Time,10)
              End;
          #75: Push := -2;
          #77: Push := +2;
          Else Dummy := ReadKey;
       End (* Case *)
      End; (* Then *)
    Inc (I); Delay (Time);
    Inc(J);

    If I >= MaxCredit Then SuicidePhase := False;
    If KillPhase and (I > 2 * MaxCredit)
      Then Begin
       KillPhase := False;
       MFSLearn := True;
       J := 1; ErrorAll := 0;
      End;

    If (Abs (GetInState ('Angle')) > 45.0)            (* Wenn umgefallen *)
      Then Begin
       SetInState(SystemInState, 'Angle', 0);
       SetInState(OldSystemInState, 'Angle', 0);
       StartWG := -0.1 * (StartWG / Abs(StartWG));
       SetInState(SystemInState, 'AngleVelocity', StartWG);
       SetInState(OldSystemInState, 'AngleVelocity', StartWG);
       SystemOutState^.ChangeState(0);
       OldSystemOutState^.ChangeState(0);
       NewTrial := True;
       Inc (Starts);
      End;
   End; (* While *)

  Window (1,1,80,25);
  GotoXY (27,25); Write ('ENTER = Weiter, ESC = Exit');
  Repeat
   Dummy := ReadKey;
  Until Dummy in [#13,#27];
  Exit := Dummy = #27;
  ClrScr;
  If Logging Then Close (Log);
  If Learning Then Begin
   Writeln (Result, 'Zustand nach einem Lernvorgang mit den Parametern');
   Writeln (Result, 'Winkel: ', StartW:0:2, ', Winkelgeschw.: ',
            StartWG:0:2, ', Loops: ', Loops, ', Versuche: ', Starts);
   Writeln (Result);
   PrintFuzzyVariables (Result)
  End;

 Until Exit;

 Print_All_Rules (Controller, Rulefile);
 PrintRuleMatrix (Controller, Rulefile);

 Controller^.Done;
 Pendulum^.Done;
 Close (Result);
 Close (Rulefile);
 Close (KillLog);

End.